@@ -3,6 +3,8 @@
 #include <assert.h>
 #include <stdlib.h>
 
+#include <openssl/crypto.h>
+
 #include "attrs.h"
 #include "log.h"
 #include "pkcs11.h"
@@ -15,15 +17,12 @@ struct attr_list {
     CK_ATTRIBUTE_PTR attrs;
 };
 
-typedef bool (*typefn)(CK_ATTRIBUTE_PTR a, CK_BYTE type, attr_list *l);
-
-#define ADD_ATTR_HANDLER(t, m) { .type = t, .tfn = add_type_copy, .memtype = m }
+#define ADD_ATTR_HANDLER(t, m) { .type = t, .memtype = m }
 
 typedef struct attr_handler2 attr_handler2;
 struct attr_handler2 {
     CK_ATTRIBUTE_TYPE type;
     CK_BYTE memtype;
-    typefn tfn;
 };
 
 #define ALLOC_LEN 16
@@ -41,11 +40,7 @@ static bool _attr_list_add(attr_list *l,
         }
 
         size_t bytes = 0;
-        res = __builtin_mul_overflow(l->max, sizeof(*l->attrs), &bytes);
-        if (res) {
-            LOGE("mul overflow\n");
-            return false;
-        }
+        safe_mul(bytes, l->max, sizeof(*l->attrs));
 
         void *tmp = realloc(l->attrs, bytes);
         if (!tmp) {
@@ -60,7 +55,8 @@ static bool _attr_list_add(attr_list *l,
          * If the mul operation didn't overflow above, then
          * mul cannot overflow here, so use regular mul
          */
-        memset(&l->attrs[l->count], 0, sizeof(*l->attrs) * ALLOC_LEN);
+        safe_mul(bytes, ALLOC_LEN, sizeof(*l->attrs));
+        memset(&l->attrs[l->count], 0, bytes);
     }
 
     /* only hex strings and sequences can be empty */
@@ -121,7 +117,9 @@ static attr_handler2 attr_handlers[] = {
     ADD_ATTR_HANDLER(CKA_TOKEN, TYPE_BYTE_BOOL),
     ADD_ATTR_HANDLER(CKA_PRIVATE, TYPE_BYTE_BOOL),
     ADD_ATTR_HANDLER(CKA_LABEL, TYPE_BYTE_HEX_STR),
+    ADD_ATTR_HANDLER(CKA_APPLICATION, TYPE_BYTE_HEX_STR),
     ADD_ATTR_HANDLER(CKA_VALUE, TYPE_BYTE_HEX_STR),
+    ADD_ATTR_HANDLER(CKA_OBJECT_ID, TYPE_BYTE_HEX_STR),
     ADD_ATTR_HANDLER(CKA_CERTIFICATE_TYPE, TYPE_BYTE_INT),
     ADD_ATTR_HANDLER(CKA_ISSUER, TYPE_BYTE_HEX_STR),
     ADD_ATTR_HANDLER(CKA_SERIAL_NUMBER, TYPE_BYTE_HEX_STR),
@@ -173,7 +171,7 @@ static attr_handler2 attr_handlers[] = {
     ADD_ATTR_HANDLER(CKA_TPM2_PRIV_BLOB, TYPE_BYTE_HEX_STR),
 };
 
-static attr_handler2 default_handler = { .tfn = add_type_copy, .memtype = 0 };
+static attr_handler2 default_handler = { .memtype = 0 };
 
 static attr_handler2 *attr_lookup(CK_ATTRIBUTE_TYPE t) {
 
@@ -221,6 +219,16 @@ CK_ATTRIBUTE_PTR attr_list_get_ptr(attr_list *l) {
     return l->attrs;
 }
 
+void attr_pfree_cleanse(CK_ATTRIBUTE_PTR attr) {
+    if (attr && attr->pValue) {
+        /* Don't use clear_free because OSSL is not used for alloc */
+        OPENSSL_cleanse(attr->pValue, attr->ulValueLen);
+        free(attr->pValue);
+        attr->pValue = NULL;
+        attr->ulValueLen = 0;
+    }
+}
+
 void attr_list_free(attr_list *attrs) {
 
     if (!attrs) {
@@ -230,7 +238,7 @@ void attr_list_free(attr_list *attrs) {
     CK_ULONG i;
     for (i=0; i < attrs->count; i++) {
         const CK_ATTRIBUTE_PTR a = &attrs->attrs[i];
-        free(a->pValue);
+        attr_pfree_cleanse(a);
     }
 
     free(attrs->attrs);
@@ -285,7 +293,7 @@ bool attr_typify(CK_ATTRIBUTE_PTR attrs, CK_ULONG cnt, attr_list **copy) {
     for (i=0; i < cnt; i++) {
         CK_ATTRIBUTE_PTR a = &attrs[i];
         attr_handler2 *h = attr_lookup(a->type);
-        bool res = h->tfn(a, h->memtype, c);
+        bool res = add_type_copy(a, h->memtype, c);
         if (!res) {
             attr_list_free(c);
             return res;
@@ -297,6 +305,55 @@ bool attr_typify(CK_ATTRIBUTE_PTR attrs, CK_ULONG cnt, attr_list **copy) {
     return true;
 }
 
+CK_RV attr_list_dup(attr_list *old, attr_list **new) {
+    assert(old);
+    assert(new);
+
+    CK_RV rv = CKR_GENERAL_ERROR;
+
+    /* create the container */
+    attr_list *tmp = calloc(1, sizeof(attr_list));
+    if (!tmp) {
+        LOGE("oom");
+        return CKR_HOST_MEMORY;
+    }
+
+    /* create the attribute list */
+    tmp->attrs = calloc(old->max, sizeof(CK_ATTRIBUTE));
+    if (!tmp->attrs) {
+        LOGE("oom");
+        free(tmp);
+        return CKR_HOST_MEMORY;
+    }
+    tmp->max = old->max;
+
+    /* deep copy the attrs */
+    size_t i;
+    for (i=0; i < old->count; i++) {
+        CK_ATTRIBUTE_PTR o = &old->attrs[i];
+        CK_ATTRIBUTE_PTR n = &tmp->attrs[i];
+
+        n->type = o->type;
+        if (o->pValue && o->ulValueLen) {
+            rv = type_mem_dup(o->pValue, o->ulValueLen, &n->pValue);
+            if (rv != CKR_OK) {
+                goto error;
+            }
+            n->ulValueLen = o->ulValueLen;
+        }
+
+        tmp->count++;
+    }
+
+    *new = tmp;
+
+    return CKR_OK;
+
+error:
+    attr_list_free(tmp);
+    return rv;
+}
+
 CK_ATTRIBUTE_PTR attr_get_attribute_by_type_raw(CK_ATTRIBUTE_PTR haystack, CK_ULONG haystack_count,
         CK_ATTRIBUTE_TYPE needle) {
 
@@ -336,7 +393,9 @@ attr_list *attr_list_append_attrs(
     /* todo safe addition */
     CK_ULONG old_len = attr_list_get_count(old_attrs);
     CK_ULONG new_len = attr_list_get_count(*new_attrs);
-    CK_ULONG total_len = new_len + old_len;
+
+    CK_ULONG total_len = 0;
+    safe_add(total_len, new_len, old_len);
 
     if (!new_len) {
         attr_list_free(*new_attrs);
@@ -348,10 +407,14 @@ attr_list *attr_list_append_attrs(
     if (total_len > old_attrs->max) {
 
         size_t blocks = total_len / ALLOC_LEN;
-        blocks += total_len % ALLOC_LEN ? 1 : 0;
+        safe_adde(blocks, total_len % ALLOC_LEN ? 1 : 0);
 
-        CK_ULONG alloc_items = blocks * ALLOC_LEN;
-        void *tmp = realloc(old_attrs->attrs, alloc_items * sizeof(CK_ATTRIBUTE));
+        CK_ULONG alloc_items = 0;
+        safe_mul(alloc_items, blocks, ALLOC_LEN);
+
+        size_t bytes = 0;
+        safe_mul(bytes, alloc_items, sizeof(CK_ATTRIBUTE));
+        void *tmp = realloc(old_attrs->attrs, bytes);
         if (!tmp) {
             return NULL;
         }
@@ -360,12 +423,17 @@ attr_list *attr_list_append_attrs(
 
         /* clear the delta */
         size_t delta = alloc_items - old_attrs->max;
-        memset(clear_point, 0, delta * sizeof(CK_ATTRIBUTE));
+        safe_mul(bytes, delta, sizeof(CK_ATTRIBUTE));
+        memset(clear_point, 0, bytes);
         old_attrs->max = alloc_items;
     }
 
     CK_ATTRIBUTE_PTR cpy_point = &old_attrs->attrs[old_len];
-    memcpy(cpy_point, (*new_attrs)->attrs, new_len * sizeof(CK_ATTRIBUTE));
+
+    size_t bytes = 0;
+    safe_mul(bytes, new_len,  sizeof(CK_ATTRIBUTE));
+
+    memcpy(cpy_point, (*new_attrs)->attrs, bytes);
 
     old_attrs->count = total_len;
 
@@ -376,22 +444,16 @@ attr_list *attr_list_append_attrs(
     return old_attrs;
 }
 
-static CK_RV attr_common_add_storage(attr_list **storage_attrs) {
+CK_RV attr_common_add_storage(attr_list **storage_attrs) {
 
     CK_RV rv = CKR_GENERAL_ERROR;
 
-    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(*storage_attrs, CKA_CLASS);
-    if (!a) {
+    CK_OBJECT_CLASS cka_class = attr_list_get_CKA_CLASS(*storage_attrs, CK_OBJECT_CLASS_BAD);
+    if (cka_class == CK_OBJECT_CLASS_BAD) {
         LOGE("Expected object to have CKA_CLASS");
         return CKR_GENERAL_ERROR;
     }
 
-    CK_ULONG v;
-    rv = attr_CK_ULONG(a, &v);
-    if (rv != CKR_OK) {
-        return rv;
-    }
-
     attr_list *new_attrs = attr_list_new();
     if (!new_attrs) {
         LOGE("oom");
@@ -401,26 +463,28 @@ static CK_RV attr_common_add_storage(attr_list **storage_attrs) {
     /* currently immutable */
     bool r = attr_list_add_bool(new_attrs, CKA_TOKEN, CK_TRUE);
     goto_error_false(r);
-    attr_list_add_bool(new_attrs, CKA_MODIFIABLE, CK_FALSE);
+    r = attr_list_add_bool(new_attrs, CKA_MODIFIABLE, CK_TRUE);
     goto_error_false(r);
-    attr_list_add_bool(new_attrs, CKA_COPYABLE, CK_TRUE);
+    r = attr_list_add_bool(new_attrs, CKA_COPYABLE, CK_TRUE);
     goto_error_false(r);
-    attr_list_add_bool(new_attrs, CKA_DESTROYABLE, CK_FALSE);
+    r = attr_list_add_bool(new_attrs, CKA_DESTROYABLE, CK_TRUE);
     goto_error_false(r);
 
     /* defaults */
-    CK_BBOOL defpriv = ((v == CKO_PRIVATE_KEY) || (v == CKO_SECRET_KEY)) ?
-            CK_TRUE : CK_FALSE;
+    CK_BBOOL defpriv = ((cka_class == CKO_PRIVATE_KEY) ||
+            (cka_class == CKO_SECRET_KEY)              ||
+            (cka_class == CKO_DATA)) ?
+                CK_TRUE : CK_FALSE;
 
-    a = attr_get_attribute_by_type(*storage_attrs, CKA_PRIVATE);
+    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(*storage_attrs, CKA_PRIVATE);
     if (!a) {
-        attr_list_add_bool(new_attrs, CKA_PRIVATE, defpriv);
+        r = attr_list_add_bool(new_attrs, CKA_PRIVATE, defpriv);
         goto_error_false(r);
     }
 
     a = attr_get_attribute_by_type(*storage_attrs, CKA_LABEL);
     if (!a) {
-        attr_list_add_buf(new_attrs, CKA_LABEL, NULL, 0);
+        r = attr_list_add_buf(new_attrs, CKA_LABEL, NULL, 0);
         goto_error_false(r);
     }
 
@@ -435,6 +499,51 @@ error:
     return rv;
 }
 
+CK_RV attr_common_add_data(attr_list **data_attrs) {
+
+    CK_RV rv = CKR_HOST_MEMORY;
+
+    /* expected */
+    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(*data_attrs, CKA_VALUE);
+    if (!a) {
+        LOGE("Expected object to have CKA_VALUE");
+        return CKR_TEMPLATE_INCOMPLETE;
+    } else  if (!a->ulValueLen || !a->pValue) {
+        LOGE("CKA_VALUE bad, got len: %lu, pValue: %s", a->ulValueLen,
+                a->pValue ? "(set)" : "(null)");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+    }
+
+    attr_list *new_attrs = attr_list_new();
+    if (!new_attrs) {
+        LOGE("oom");
+        return CKR_HOST_MEMORY;
+    }
+
+    a = attr_get_attribute_by_type(*data_attrs, CKA_OBJECT_ID);
+    if (!a) {
+        bool r = attr_list_add_buf(new_attrs, CKA_APPLICATION, NULL, 0);
+        goto_error_false(r);
+    }
+
+    a = attr_get_attribute_by_type(*data_attrs, CKA_APPLICATION);
+    if (!a) {
+        bool r = attr_list_add_buf(new_attrs, CKA_APPLICATION, NULL, 0);
+        goto_error_false(r);
+    }
+
+    *data_attrs = attr_list_append_attrs(*data_attrs,
+            &new_attrs);
+    goto_error_false(*data_attrs);
+
+    return attr_common_add_storage(data_attrs);
+
+error:
+    attr_list_free(new_attrs);
+
+    return rv;
+}
+
 static CK_RV attr_common_add_key(attr_list **key_attrs) {
 
     CK_RV rv = CKR_HOST_MEMORY;
@@ -452,12 +561,6 @@ static CK_RV attr_common_add_key(attr_list **key_attrs) {
         return CKR_GENERAL_ERROR;
     }
 
-    a = attr_get_attribute_by_type(*key_attrs, CKA_KEY_GEN_MECHANISM);
-    if (!a) {
-        LOGE("Expected object to have CKA_KEY_GEN_MECHANISM");
-        return CKR_GENERAL_ERROR;
-    }
-
     a = attr_get_attribute_by_type(*key_attrs, CKA_ALLOWED_MECHANISMS);
     if (!a) {
         LOGE("Expected object to have CKA_ALLOWED_MECHANISMS");
@@ -581,10 +684,12 @@ CK_RV attr_common_add_RSA_publickey(attr_list **public_attrs) {
     if (!a) {
         a = attr_get_attribute_by_type(*public_attrs, CKA_MODULUS);
         if (!a) {
-            LOGE("Expeted object to have CKA_MODULUS");
+            LOGE("Expected object to have CKA_MODULUS");
             goto error;
         }
-        CK_ULONG modulus_bits = a->ulValueLen * 8;
+
+        CK_ULONG modulus_bits = 0;
+        safe_mul(modulus_bits, a->ulValueLen, 8);
         bool r = attr_list_add_int(new_attrs, CKA_MODULUS_BITS, modulus_bits);
         goto_error_false(r);
     }
@@ -735,7 +840,8 @@ static CK_RV attr_common_add_RSA_privatekey(attr_list **private_attrs) {
         a = attr_get_attribute_by_type(*private_attrs, CKA_MODULUS);
         /* we checked above this cant fail */
         assert(a);
-        CK_ULONG modulus_bits = a->ulValueLen * 8;
+        CK_ULONG modulus_bits = 0;
+        safe_mul(modulus_bits, a->ulValueLen, 8);
         bool r = attr_list_add_int(new_attrs, CKA_MODULUS_BITS, modulus_bits);
         goto_error_false(r);
     }
@@ -774,7 +880,11 @@ CK_RV rsa_gen_mechs(attr_list *new_pub_attrs, attr_list *new_priv_attrs) {
         CKM_SHA256_RSA_PKCS,
         CKM_SHA384_RSA_PKCS,
         CKM_SHA512_RSA_PKCS,
-        CKM_RSA_PKCS_PSS
+        CKM_RSA_PKCS_PSS,
+        CKM_SHA1_RSA_PKCS_PSS,
+        CKM_SHA256_RSA_PKCS_PSS,
+        CKM_SHA384_RSA_PKCS_PSS,
+        CKM_SHA512_RSA_PKCS_PSS,
     };
 
     if (new_pub_attrs) {
@@ -1001,7 +1111,7 @@ add_item:
         /* no - add it, shallow copy ok */
         h = attr_lookup(cur->type);
         assert(h);
-        r = h->tfn(cur, h->memtype, d);
+        r = add_type_copy(cur, h->memtype, d);
         if (!r) {
             attr_list_free(d);
             return CKR_GENERAL_ERROR;
@@ -1102,6 +1212,106 @@ error:
     return rv;
 }
 
+CK_RV attr_list_update_entry(attr_list *attrs, CK_ATTRIBUTE_PTR untrusted_attr) {
+    assert(attrs);
+    assert(untrusted_attr);
+
+    CK_ATTRIBUTE_TYPE t = untrusted_attr->type;
+
+    attr_handler2 *handler = attr_lookup(t);
+
+    CK_ATTRIBUTE_PTR found = attr_get_attribute_by_type(attrs, t);
+    if (!found) {
+        LOGE("Attribute entry not found");
+        return CKR_GENERAL_ERROR;
+    }
+
+    /* the existing type memory should match the expected type memory */
+    CK_BYTE expected_memory_type = type_from_ptr(found->pValue,
+            found->ulValueLen);
+
+    /* internal state check */
+    if (expected_memory_type != handler->memtype) {
+        LOGE("expected memory(%u-%s) != handler memory(%u-%s)",
+            expected_memory_type, type_to_str(expected_memory_type),
+            handler->memtype, type_to_str(handler->memtype));
+        return CKR_GENERAL_ERROR;
+    }
+
+    /* validate sizes */
+    void *pValue = untrusted_attr->pValue;
+    CK_ULONG ulValueLen = untrusted_attr->ulValueLen;
+
+    switch (handler->memtype) {
+    case TYPE_BYTE_INT:
+        if (ulValueLen != sizeof(CK_ULONG)) {
+            LOGE("ulValueLen(%lu) != sizeof(CK_ULONG)", ulValueLen);
+            return CKR_ATTRIBUTE_VALUE_INVALID;
+        }
+        break;
+    case TYPE_BYTE_BOOL:
+        if (ulValueLen != sizeof(CK_BBOOL)) {
+            LOGE("ulValueLen(%lu) != sizeof(CK_BBOOL)", ulValueLen);
+            return CKR_ATTRIBUTE_VALUE_INVALID;
+        }
+        break;
+    case TYPE_BYTE_INT_SEQ:
+        if (ulValueLen % sizeof(CK_ULONG)) {
+            LOGE("ulValueLen(%lu) %% sizeof(CK_ULONG)",
+                    ulValueLen % sizeof(CK_ULONG));
+            return CKR_ATTRIBUTE_VALUE_INVALID;
+        }
+        break;
+    case TYPE_BYTE_HEX_STR:
+        /* nothing to do */
+        break;
+    default:
+        LOGE("Unknown data type representation, got: %u",
+                handler->memtype);
+        return CKR_GENERAL_ERROR;
+    }
+
+    if (ulValueLen != found->ulValueLen) {
+        void *new_pValue = type_zrealloc(found->pValue, ulValueLen, handler->memtype);
+        if (!new_pValue) {
+            LOGE("oom");
+            return CKR_HOST_MEMORY;
+        }
+        /* update the found node with the new resized memory */
+        found->ulValueLen = ulValueLen;
+        found->pValue = new_pValue;
+    }
+
+    /* update the contents of memory with what was provided */
+    memcpy(found->pValue, pValue, ulValueLen);
+
+    return CKR_OK;
+}
+
+CK_RV attr_list_append_entry(attr_list **attrs, CK_ATTRIBUTE_PTR untrusted_attr) {
+    assert(attrs);
+    assert(*attrs);
+    assert(untrusted_attr);
+
+    attr_list *new_item = NULL;
+    bool res = attr_typify(untrusted_attr, 1, &new_item);
+    if (!res) {
+        LOGE("Could not typify attr: %lu", untrusted_attr->type);
+        return CKR_GENERAL_ERROR;
+    }
+
+    attr_list *x = attr_list_append_attrs(
+            *attrs,
+            &new_item);
+    if (!x) {
+        return CKR_GENERAL_ERROR;
+    }
+
+    *attrs = x;
+
+    return CKR_OK;
+}
+
 #define UTILS_GENERIC_ATTR_TYPE_CONVERT(T) \
     CK_RV attr_##T(CK_ATTRIBUTE_PTR attr, T *x) { \
       assert(attr); \
@@ -1120,3 +1330,18 @@ UTILS_GENERIC_ATTR_TYPE_CONVERT(CK_ULONG);
 UTILS_GENERIC_ATTR_TYPE_CONVERT(CK_BBOOL);
 UTILS_GENERIC_ATTR_TYPE_CONVERT(CK_OBJECT_CLASS);
 UTILS_GENERIC_ATTR_TYPE_CONVERT(CK_KEY_TYPE);
+
+#define UTILS_GENERIC_EXTRACTOR(type, T) \
+    T attr_list_get_##type(attr_list *attrs, T defvalue) { \
+	    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(attrs, type); \
+	    if (!a) { \
+	        return defvalue; \
+	    } \
+	    T x = defvalue; \
+		CK_RV rv = attr_##T(a, &x); \
+		return rv != CKR_OK ? defvalue : x; \
+    }
+
+UTILS_GENERIC_EXTRACTOR(CKA_PRIVATE, CK_BBOOL);
+UTILS_GENERIC_EXTRACTOR(CKA_CLASS,   CK_OBJECT_CLASS);
+UTILS_GENERIC_EXTRACTOR(CKA_TOKEN, CK_BBOOL);
