@@ -870,6 +870,191 @@ static void test_non_common_template_attrs(void **state) {
     assert_int_equal(rv, CKR_OK);
 }
 
+static void test_extract_and_private(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    CK_BBOOL ck_true = CK_TRUE;
+    // CK_BBOOL ck_false = CK_FALSE;
+    CK_UTF8CHAR label[] = "minimum-rsa";
+
+    CK_ATTRIBUTE pub[] = {
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_ENCRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_VERIFY, ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+    };
+
+    CK_ATTRIBUTE priv[] = {
+        ADD_ATTR_BASE(CKA_DECRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_SIGN, ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+        ADD_ATTR_BASE(CKA_EXTRACTABLE, ck_true),
+    };
+
+    CK_MECHANISM mech = {
+        .mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN,
+        .pParameter = NULL,
+        .ulParameterLen = 0
+    };
+
+    CK_OBJECT_HANDLE pubkey;
+    CK_OBJECT_HANDLE privkey;
+
+    user_login(session);
+
+    CK_RV rv = C_GenerateKeyPair (session,
+            &mech,
+            pub, ARRAY_LEN(pub),
+            priv, ARRAY_LEN(priv),
+            &pubkey, &privkey);
+    assert_int_equal(rv, CKR_ATTRIBUTE_VALUE_INVALID);
+}
+
+static void test_rsa_keypairgen_wrap(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    CK_BBOOL ck_true = CK_TRUE;
+    // CK_BBOOL ck_false = CK_FALSE;
+    CK_UTF8CHAR label[] = "wrap-rsa";
+
+    CK_ATTRIBUTE pub[] = {
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_ENCRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_VERIFY, ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+        ADD_ATTR_BASE(CKA_WRAP, ck_true),
+    };
+
+    CK_ATTRIBUTE priv[] = {
+        ADD_ATTR_BASE(CKA_DECRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_SIGN, ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+        ADD_ATTR_BASE(CKA_EXTRACTABLE, ck_true),
+    };
+
+    CK_MECHANISM mech = {
+        .mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN,
+        .pParameter = NULL,
+        .ulParameterLen = 0
+    };
+
+    CK_OBJECT_HANDLE pubkey;
+    CK_OBJECT_HANDLE privkey;
+
+    user_login(session);
+
+    CK_RV rv = C_GenerateKeyPair (session,
+            &mech,
+            pub, ARRAY_LEN(pub),
+            priv, ARRAY_LEN(priv),
+            &pubkey, &privkey);
+    assert_int_equal(rv, CKR_ATTRIBUTE_VALUE_INVALID);
+}
+
+static void test_rsa_keypairgen_unwrap(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    CK_BBOOL ck_true = CK_TRUE;
+    // CK_BBOOL ck_false = CK_FALSE;
+    CK_UTF8CHAR label[] = "unwrap-rsa";
+
+    CK_ATTRIBUTE pub[] = {
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_ENCRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_VERIFY, ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+    };
+
+    CK_ATTRIBUTE priv[] = {
+        ADD_ATTR_BASE(CKA_DECRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_SIGN, ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+        ADD_ATTR_BASE(CKA_EXTRACTABLE, ck_true),
+        ADD_ATTR_BASE(CKA_UNWRAP, ck_true),
+    };
+
+    CK_MECHANISM mech = {
+        .mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN,
+        .pParameter = NULL,
+        .ulParameterLen = 0
+    };
+
+    CK_OBJECT_HANDLE pubkey;
+    CK_OBJECT_HANDLE privkey;
+
+    user_login(session);
+
+    CK_RV rv = C_GenerateKeyPair (session,
+            &mech,
+            pub, ARRAY_LEN(pub),
+            priv, ARRAY_LEN(priv),
+            &pubkey, &privkey);
+    assert_int_equal(rv, CKR_ATTRIBUTE_VALUE_INVALID);
+}
+
+static void test_rsa_keypairgen_wrap_unwrap(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    CK_BBOOL ck_true = CK_TRUE;
+    // CK_BBOOL ck_false = CK_FALSE;
+    CK_UTF8CHAR label[] = "wrap-unwrap-rsa";
+
+    CK_ATTRIBUTE pub[] = {
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_ENCRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_VERIFY, ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+        ADD_ATTR_BASE(CKA_WRAP, ck_true),
+    };
+
+    CK_ATTRIBUTE priv[] = {
+        ADD_ATTR_BASE(CKA_DECRYPT, ck_true),
+        ADD_ATTR_BASE(CKA_SIGN, ck_true),
+        ADD_ATTR_BASE(CKA_PRIVATE, ck_true),
+        ADD_ATTR_BASE(CKA_TOKEN,   ck_true),
+        ADD_ATTR_STR(CKA_LABEL, label),
+        ADD_ATTR_BASE(CKA_EXTRACTABLE, ck_true),
+        ADD_ATTR_BASE(CKA_UNWRAP, ck_true),
+    };
+
+    CK_MECHANISM mech = {
+        .mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN,
+        .pParameter = NULL,
+        .ulParameterLen = 0
+    };
+
+    CK_OBJECT_HANDLE pubkey;
+    CK_OBJECT_HANDLE privkey;
+
+    user_login(session);
+
+    CK_RV rv = C_GenerateKeyPair (session,
+            &mech,
+            pub, ARRAY_LEN(pub),
+            priv, ARRAY_LEN(priv),
+            &pubkey, &privkey);
+    assert_int_equal(rv, CKR_ATTRIBUTE_VALUE_INVALID);
+}
+
+
 static void test_create_obj_rsa_public_key(void **state) {
 
     test_info *ti = test_info_from_state(state);
@@ -946,45 +1131,119 @@ static void test_create_obj_rsa_public_key(void **state) {
     rv = C_Decrypt(session, ciphertext, ciphertext_len, plaintext2, &plaintext2_len);
     assert_int_equal(rv, CKR_OK);
 
-    /* mode has the plaintext at end, so strip */
-    CK_BYTE *p2 = &plaintext2[sizeof(plaintext2) - sizeof(plaintext)];
+    assert_int_equal(plaintext2_len, sizeof(plaintext));
+    assert_memory_equal(plaintext2, plaintext, sizeof(plaintext));
+}
 
-    assert_memory_equal(plaintext, p2, sizeof(plaintext));
+/*
+ * Place this test here so we don't have to take the test time hit of another
+ * test setup.
+ */
+static void test_create_data_object_private (void **state) {
 
-    /*
-     * FIXME #384
-     * Encrypt with priv key and decrypt with pubkey
-     *
-     * This doesn't seem to be working as expected, and it likely
-     * this issue:
-     *   - https://superuser.com/questions/1363466/how-to-decrypt-with-public-key-after-encrypt-with-private-key
-     *
-     * Considering that this method isn't a RAW rsa operation, the TPM is likely using the public portion.
-     * So this may never work with this mode.
-     *
-     * Currently fails with:
-     *  ERROR on line: "203" in file: "src/lib/encrypt.c": Could not perform RSA public decrypt: error:0407008A:rsa routines:RSA_padding_check_PKCS1_type_1:invalid padding
-     */
-//    rv = C_EncryptInit(session, &mech, privkey);
-//    assert_int_equal(rv, CKR_OK);
-//
-//    rv = C_Encrypt(session, plaintext, sizeof(plaintext), ciphertext, &ciphertext_len);
-//    assert_int_equal(rv, CKR_OK);
-//
-//    rv = C_DecryptInit(session, &mech, pubkey);
-//    assert_int_equal(rv, CKR_OK);
-//
-//    rv = C_Decrypt(session, ciphertext, ciphertext_len, plaintext2, &plaintext2_len);
-//    assert_int_equal(rv, CKR_OK);
-//
-//    p2 = &plaintext2[sizeof(plaintext2) - sizeof(plaintext)];
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
 
-    assert_memory_equal(plaintext, p2, sizeof(plaintext));
+    user_login(session);
+
+    CK_BBOOL _false = CK_FALSE;
+    CK_BBOOL _true = CK_TRUE;
+
+    char label[] = "data object";
+    char application[] = "my application";
+
+    CK_OBJECT_CLASS object_class = CKO_DATA;
+
+    CK_BYTE id[] = { '1', '2', '3' };
+    CK_BYTE value[] = "my data object";
+
+    CK_ATTRIBUTE data_template[] = {
+      { CKA_CLASS,       &object_class, sizeof(object_class)    },
+      { CKA_TOKEN,       &_true,        sizeof(_true)           },
+      { CKA_PRIVATE,     &_true,        sizeof(_true)           },
+      { CKA_LABEL,       label,         sizeof(label) - 1       },
+      { CKA_MODIFIABLE,  &_false,       sizeof(_false)          },
+      { CKA_APPLICATION, &application,  sizeof(application) - 1 },
+      { CKA_OBJECT_ID,   &id,           sizeof(id)              },
+      { CKA_VALUE,       value,         sizeof(value)           }
+    };
+
+    CK_OBJECT_HANDLE obj = 0;
+
+    CK_RV rv = C_CreateObject(session, data_template, ARRAY_LEN(data_template), &obj);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_BYTE buf[64];
+
+    CK_ATTRIBUTE data_template2[] = {
+      { CKA_VALUE, buf, sizeof(buf) },
+    };
+
+    rv = C_GetAttributeValue (session, obj,
+            data_template2, ARRAY_LEN(data_template2));
+    assert_int_equal(rv, CKR_OK);
+
+    assert_int_equal(sizeof(value), data_template2[0].ulValueLen);
+    assert_memory_equal(value, buf, sizeof(value));
+
+    logout(session);
+
+    CK_BYTE buf2[64] = { 0 };
+    CK_ATTRIBUTE data_template3[] = {
+      { CKA_VALUE, buf2, sizeof(buf2) },
+    };
+
+    rv = C_GetAttributeValue (session, obj,
+            data_template3, ARRAY_LEN(data_template3));
+    assert_int_equal(rv, CKR_OK);
+
+    assert_int_equal(data_template3[0].ulValueLen, 0);
+}
+
+static void test_create_data_object_public (void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_BBOOL _false = CK_FALSE;
+    CK_BBOOL _true = CK_TRUE;
+
+    char label[] = "public data object";
+    char application[] = "my application";
+
+    CK_OBJECT_CLASS object_class = CKO_DATA;
+
+    CK_BYTE id[] = { '1', '2', '3', '4' };
+    CK_BYTE value[] = "my public data object";
+
+    CK_ATTRIBUTE data_template[] = {
+      { CKA_CLASS,       &object_class, sizeof(object_class)    },
+      { CKA_TOKEN,       &_true,        sizeof(_true)           },
+      { CKA_PRIVATE,     &_false,        sizeof(_false)          },
+      { CKA_LABEL,       label,         sizeof(label) - 1       },
+      { CKA_MODIFIABLE,  &_false,       sizeof(_false)          },
+      { CKA_APPLICATION, &application,  sizeof(application) - 1 },
+      { CKA_OBJECT_ID,   &id,           sizeof(id)              },
+      { CKA_VALUE,       value,         sizeof(value)           }
+    };
+
+    CK_OBJECT_HANDLE obj = 0;
+
+    CK_RV rv = C_CreateObject(session, data_template, ARRAY_LEN(data_template), &obj);
+    assert_int_equal(rv, CKR_ATTRIBUTE_VALUE_INVALID);
 }
 
 int main() {
 
     const struct CMUnitTest tests[] = {
+        cmocka_unit_test_setup_teardown(test_extract_and_private,
+                test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_create_data_object_public,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_create_data_object_private,
+            test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_create_obj_rsa_public_key,
             test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_rsa_keygen_missing_attributes,
@@ -1001,6 +1260,12 @@ int main() {
             test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_non_common_template_attrs,
             test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_rsa_keypairgen_wrap,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_rsa_keypairgen_unwrap,
+                test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_rsa_keypairgen_wrap_unwrap,
+                test_setup, test_teardown),
     };
 
     return cmocka_run_group_tests(tests, group_setup, group_teardown);
