@@ -3,16 +3,18 @@
 import binascii
 import io
 import os
+import struct
 import sys
 import yaml
 
+from tempfile import mkstemp
+
 # local imports
 from .command import Command
 from .command import commandlet
 from .db import Db
 from .objects import PKCS11ObjectFactory as PKCS11ObjectFactory
 from .objects import PKCS11X509
-from .utils import bytes_to_file
 from .utils import AESAuthUnwrapper
 from .utils import TemporaryDirectory
 from .utils import hash_pass
@@ -20,6 +22,8 @@ from .utils import rand_hex_str
 from .utils import pemcert_to_attrs
 from .utils import str2bool
 from .utils import str2bytes
+from .utils import asn1parse_tss_key
+from .utils import get_pobject
 
 from .tpm2 import Tpm2
 
@@ -31,6 +35,13 @@ class NewKeyCommandBase(Command):
     '''
 
     def generate_options(self, group_parser):
+        group_parser.add_argument(
+            '--label',
+            help='The tokens label to import the key too.\n',
+            required=True)
+        group_parser.add_argument(
+            '--key-label',
+            help='The label of the key imported. Defaults to an integer value.\n')
         group_parser.add_argument(
             '--id',
             help='The key id. Defaults to a random 8 bytes of hex.\n',
@@ -39,18 +50,21 @@ class NewKeyCommandBase(Command):
             '--attr-always-authenticate',
             action='store_true',
             help='Sets the CKA_ALWAYS_AUTHENTICATE attribute to CK_TRUE.\n')
+        group_parser.add_argument(
+            '--hierarchy-auth',
+            help='The hierarchyauth, required for transient pobjects.\n',
+            default='')
         pinopts = group_parser.add_mutually_exclusive_group(required=True)
         pinopts.add_argument('--sopin', help='The Administrator pin.\n'),
         pinopts.add_argument('--userpin', help='The User pin.\n'),
 
     # Implemented by derived class
-    def new_key_create(self, pobj, objauth, tpm2, path, alg, privkey, d):
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, d):
         raise NotImplementedError('Implement: new_key')
 
-    @staticmethod
-    def new_key_init(label, sopin, userpin, pobj, sealobjects, tpm2, d):
+    def new_key_init(self, label, sopin, userpin, hierarchyauth, pobj, sealobjects, tpm2, d):
 
-        tr_handle = bytes_to_file(pobj['handle'], d)
+        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
 
         # Get the primary object encrypted auth value and sokey information
         # to decode it. Based on the incoming pin
@@ -65,7 +79,7 @@ class NewKeyCommandBase(Command):
         sealpriv = sealobjects[privkey]
         sealsalt = sealobjects[saltkey]
 
-        sealctx = tpm2.load(tr_handle, pobj['objauth'], sealpriv, sealpub)
+        sealctx = tpm2.load(pobj_handle, pobj['objauth'], sealpriv, sealpub)
 
         sealauth = hash_pass(pin, salt=sealsalt)['hash']
 
@@ -149,9 +163,10 @@ class NewKeyCommandBase(Command):
                 label = args['label']
                 sopin = args['sopin']
                 userpin = args['userpin']
-                alg = args['algorithm']
+                alg = args['algorithm'] if 'algorithm' in args else None
                 key_label = args['key_label']
                 tid = args['id']
+                hierarchyauth = args['hierarchy_auth']
 
                 privkey = None
                 try:
@@ -159,19 +174,18 @@ class NewKeyCommandBase(Command):
                 except KeyError:
                     privkey = None
 
-                path = args['path']
-
                 token = db.gettoken(label)
                 pobjectid = token['pid']
                 pobj = db.getprimary(pobjectid)
 
                 sealobjects = db.getsealobject(token['id'])
 
-                encobjauth, objauth = NewKeyCommandBase.new_key_init(
-                    label, sopin, userpin, pobj, sealobjects, tpm2, d)
+                encobjauth, objauth = self.new_key_init(
+                    label, sopin, userpin, hierarchyauth,
+                    pobj, sealobjects, tpm2, d)
 
                 tertiarypriv, tertiarypub, tertiarypubdata = self.new_key_create(
-                    pobj, objauth, tpm2, path, alg, privkey, d)
+                    pobj, objauth, hierarchyauth, tpm2, alg, privkey, d)
 
                 # handle options that can add additional attributes
                 always_auth = args['attr_always_authenticate']
@@ -196,33 +210,19 @@ class ImportCommand(NewKeyCommandBase):
             '--privkey',
             help='Full path of the private key to be imported.\n',
             required=True)
-        group_parser.add_argument(
-            '--label',
-            help='The tokens label to import the key too.\n',
-            required=True)
-        group_parser.add_argument(
-            '--key-label',
-            help='The label of the key imported. Defaults to an integer value.\n'
-        )
         group_parser.add_argument(
             '--algorithm',
             help='The type of the key.\n',
-            choices=['rsa'],
+            choices=['rsa', 'ecc'],
             required=True)
 
     # Imports a new key
-    def new_key_create(self, pobj, objauth, tpm2, path, alg, privkey, d):
-        if alg != 'rsa':
-            sys.exit('Unknown algorithm or algorithm not supported, got "%s"' %
-                     alg)
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, d):
 
-        if privkey is None:
-            sys.exit("Invalid private key path")
-
-        tr_handle = bytes_to_file(pobj['handle'], d)
+        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
 
         tertiarypriv, tertiarypub, tertiarypubdata = tpm2.importkey(
-            tr_handle, pobj['objauth'], objauth, privkey=privkey, alg=alg)
+            pobj_handle, pobj['objauth'], objauth, privkey=privkey, alg=alg)
 
         return (tertiarypriv, tertiarypub, tertiarypubdata)
 
@@ -240,30 +240,19 @@ class AddKeyCommand(NewKeyCommandBase):
     # pylint: disable=no-self-use
     def generate_options(self, group_parser):
         super(AddKeyCommand, self).generate_options(group_parser)
-        group_parser.add_argument(
-            '--label',
-            help='The tokens label to add a key too.\n',
-            required=True)
         group_parser.add_argument(
             '--algorithm',
             help='The type of the key.\n',
-            choices=[
-                'rsa1024', 'rsa2048', 'aes128', 'aes256', 'ecc224', 'ecc256',
-                'ecc384', 'ecc521'
-            ],
+            choices=Tpm2.ALGS,
             required=True)
-        group_parser.add_argument(
-            '--key-label',
-            help='The key label to identify the key. Defaults to an integer value.\n'
-        )
 
     # Creates a new key
-    def new_key_create(self, pobj, objauth, tpm2, path, alg, privkey, d):
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, d):
 
-        tr_handle = bytes_to_file(pobj['handle'], d)
+        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
 
         tertiarypriv, tertiarypub, tertiarypubdata = tpm2.create(
-            tr_handle, pobj['objauth'], objauth, alg=alg)
+            pobj_handle, pobj['objauth'], objauth, alg=alg)
 
         return (tertiarypriv, tertiarypub, tertiarypubdata)
 
@@ -411,7 +400,8 @@ class ObjMod(Command):
 
         with Db(path) as db:
             obj = db.getobject(tid)
-
+            if obj is None:
+                sys.exit('Not found, object with id: {}'.format(tid))
         s = obj['attrs']
         obj_attrs = yaml.safe_load(s)
 
@@ -525,3 +515,147 @@ class ObjDel(Command):
         path = args['path']
 
         ObjDel.delete(path, args['id'])
+
+@commandlet("link")
+class LinkCommand(NewKeyCommandBase):
+    '''
+    Imports an existing TPM key to a token within a tpm2-pkcs11 store.
+    '''
+
+    # adhere to an interface
+    # pylint: disable=no-self-use
+    def generate_options(self, group_parser):
+        super(LinkCommand, self).generate_options(group_parser)
+        group_parser.add_argument('privkey',
+            nargs='*',
+            help='Path of the key to be linked.\n')
+        group_parser.add_argument(
+            '--auth',
+            default='',
+            help='The auth value for the key to link.\n'
+        )
+
+    def new_key_init(self, label, sopin, userpin, hierarchyauth, pobj, sealobjects, tpm2, d):
+
+        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
+
+        # Get the primary object encrypted auth value and sokey information
+        # to decode it. Based on the incoming pin
+        is_so = sopin != None
+        pin = sopin if is_so else userpin
+
+        pubkey = '%spub' % ('so' if is_so else 'user')
+        privkey = '%spriv' % ('so' if is_so else 'user')
+        saltkey = '%sauthsalt' % ('so' if is_so else 'user')
+
+        sealpub = sealobjects[pubkey]
+        sealpriv = sealobjects[privkey]
+        sealsalt = sealobjects[saltkey]
+
+        sealctx = tpm2.load(pobj_handle, pobj['objauth'], sealpriv, sealpub)
+
+        sealauth = hash_pass(pin, salt=sealsalt)['hash']
+
+        wrappingkey = tpm2.unseal(sealctx, sealauth)
+
+        wrapper = AESAuthUnwrapper(wrappingkey)
+
+        objauth = self._auth
+
+        encobjauth = wrapper.wrap(str2bytes(objauth))
+
+        return (encobjauth, objauth)
+
+    def create_from_tss_key(self, pobj, objauth, hierarchyauth, tpm2, alg, keypath, d):
+
+        keypath = keypath[0]
+
+        tss2_privkey = asn1parse_tss_key(keypath)
+        is_empty_auth = bool(tss2_privkey['emptyauth'])
+        phandle = int(tss2_privkey['parent'])
+        pubbytes = bytes(tss2_privkey['pubkey'])
+        privbytes = bytes(tss2_privkey['privkey'])
+
+        pid = pobj['id']
+        pobj_config = yaml.safe_load(pobj['config'])
+        is_transient = pobj_config['transient']
+        if not is_transient and (phandle == tpm2.TPM2_RH_OWNER or \
+            (phandle >> tpm2.TPM2_HR_SHIFT != tpm2.TPM2_HT_PERSISTENT)):
+            sys.exit('The primary object (id: {:d}) is persistent and'
+                ' the TSS Engine key does not have a persistent parent,'
+                ' got: 0x{:x}'.format(pid, phandle))
+        elif is_transient and phandle != 0:
+            sys.exit('The primary object (id: {:d}) is transient and'
+                ' the TSS Engine key has a parent handle,'
+                ' got: 0x{:x}'.format(pid, phandle))
+
+
+        if is_empty_auth and len(self._auth) if self._auth is not None else 0:
+            sys.exit('Key expected to have auth value, please specify via option --auth');
+
+        if not is_transient:
+            # Im diving into the ESYS_TR serialized format,
+            # this isn't the smartest thing to do...
+            hexhandle = pobj_config['esys-tr']
+            handle_bytes = binascii.unhexlify(hexhandle)[0:4]
+            expected_handle = struct.unpack(">I", handle_bytes)[0]
+            if phandle != expected_handle:
+                sys.exit("Key must be parent of 0x{:X}, got 0x{:X}".format(
+                    expected_handle, phandle))
+
+        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
+        pobjauth = pobj['objauth']
+        ctx = tpm2.load(pobj_handle, pobjauth, privbytes, pubbytes)
+        tertiarypubdata, _ = tpm2.readpublic(ctx, False)
+
+        try:
+            privfd, tertiarypriv = mkstemp(prefix='', suffix='.priv', dir=d)
+            pubfd, tertiarypub = mkstemp(prefix='', suffix='.pub', dir=d)
+
+            os.write(privfd, privbytes)
+            os.write(pubfd, pubbytes)
+        finally:
+            os.close(privfd)
+            os.close(pubfd)
+
+        return (tertiarypriv, tertiarypub, tertiarypubdata)
+
+    def create_from_key_blobs(self, pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d):
+
+        tertiarypub = keypaths[0]
+        tertiarypriv = keypaths[1]
+
+        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
+        pobjauth = pobj['objauth']
+        try:
+            ctx = tpm2.load(pobj_handle, pobjauth, tertiarypriv, tertiarypub)
+        except RuntimeError as e:
+            # Try swapping pub/priv
+            tmp = tertiarypub
+            tertiarypub = tertiarypriv
+            tertiarypriv = tmp
+            ctx = tpm2.load(pobj_handle, pobjauth, tertiarypriv, tertiarypub)
+
+        tertiarypubdata, _ = tpm2.readpublic(ctx, False)
+
+        return (tertiarypriv, tertiarypub, tertiarypubdata)
+
+    # Links a new key
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d):
+
+        if keypaths is None:
+            sys.exit("Keypath must be specified")
+
+        if len(keypaths) == 1:
+            return self.create_from_tss_key(pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d)
+
+        if len(keypaths) == 2:
+            return self.create_from_key_blobs(pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d)
+
+        sys.exit("Expected one or two keyblobs, got: {}".format(len(keypaths)))
+
+    def __call__(self, args):
+        self._auth = args['auth'] if 'auth' in args else None
+        objects = super(LinkCommand, self).__call__(args)
+        NewKeyCommandBase.output(objects, 'link')
+
